/*
 * kgraft_patch_bsc1124734
 *
 * Fix for CVE-2019-7221, bsc#1124734
 *
 *  Upstream commit:
 *  ecec76885bcf ("KVM: nVMX: unconditionally cancel preemption timer
 *                 in free_nested (CVE-2019-7221)")
 *
 *  SLE12(-SP1) commit:
 *  2cfc10fcb868c6205c09746afa7f037b3b9a3955
 *
 *  SLE12-SP2 commit:
 *  ff1adbba7a32004dab45ee00bb3290942c593e14
 *
 *  SLE12-SP3 commit:
 *  ff1adbba7a32004dab45ee00bb3290942c593e14
 *
 *  SLE12-SP4 commit:
 *  2b60717b20c2000bbe3cca8e324f91be217b0eeb
 *
 *  SLE15 commit:
 *  2b60717b20c2000bbe3cca8e324f91be217b0eeb
 *
 *
 *  Copyright (c) 2019 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_KVM_INTEL)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kvm_host.h>
#include <linux/list.h>
#include <asm/vmx.h>
#include <linux/hrtimer.h>
#include <linux/highmem.h>
#include <linux/gfp.h>
#include <linux/slab.h>
#include <asm/msr.h>
#include <asm/special_insns.h>
#include <asm/segment.h>
#include <asm/desc.h>
#include <asm/i387.h>
#include <asm/mce.h>
#include <asm/nospec-branch.h>
#include "kgr_patch_bsc1124734.h"
#include "kallsyms_relocs.h"

#if !IS_MODULE(CONFIG_KVM_INTEL)
#error "Live patch supports only CONFIG_KVM_INTEL=m"
#endif

#define KGR_PATCHED_MODULE "kvm_intel"


struct vcpu_vmx;
struct loaded_vmcs;
struct msr_data;

static bool (*kgr_enable_shadow_vmcs);
static struct vmcs_config (*kgr_vmcs_config);
static bool (*kgr_enable_vpid);
static spinlock_t (*kgr_vmx_vpid_lock);
static DECLARE_BITMAP((*kgr_vmx_vpid_bitmap), VMX_NR_VPIDS);
static struct desc_ptr __percpu (*kgr_host_gdt);
static bool (*kgr_nested);
static const u32 (*kgr_vmx_msr_index)[];

static void (*kgr_copy_shadow_to_vmcs12)(struct vcpu_vmx *vmx);
static void (*kgr_free_loaded_vmcs)(struct loaded_vmcs *loaded_vmcs);
static int (*kgr_nested_vmx_check_permission)(struct kvm_vcpu *vcpu);
static void (*kgr_skip_emulated_instruction)(struct kvm_vcpu *vcpu);
static void (*kgr_nested_vmx_succeed)(struct kvm_vcpu *vcpu);
static void (*kgr_nested_vmx_vmexit)(struct kvm_vcpu *vcpu, u32 exit_reason,
				     u32 exit_intr_info,
				     unsigned long exit_qualification);
static void (*kgr_vmcs_writel)(unsigned long field, unsigned long value);
static bool (*kgr_guest_cpuid_has_vmx)(struct kvm_vcpu *vcpu);

static bool (*kgr_kvm_rebooting);
static struct kmem_cache *(*kgr_kvm_vcpu_cache);

static asmlinkage void (*kgr_kvm_spurious_fault)(void);
static void (*kgr_kvm_release_page_dirty)(struct page *page);
static void (*kgr_kvm_vcpu_uninit)(struct kvm_vcpu *vcpu);
static int (*kgr_kvm_set_msr_common)(struct kvm_vcpu *vcpu,
				     struct msr_data *msr_info);
static void (*kgr_kvm_write_tsc)(struct kvm_vcpu *vcpu, struct msr_data *msr);
static bool (*kgr_kvm_mtrr_valid)(struct kvm_vcpu *vcpu, u32 msr, u64 data);
static int (*kgr_kvm_set_shared_msr)(unsigned slot, u64 value, u64 mask);

static struct kgr_kallsyms_reloc kgr_funcs[] = {
	{ "enable_shadow_vmcs", (void *)&kgr_enable_shadow_vmcs, "kvm_intel" },
	{ "vmcs_config", (void *)&kgr_vmcs_config, "kvm_intel" },
	{ "enable_vpid", (void *)&kgr_enable_vpid, "kvm_intel" },
	{ "vmx_vpid_lock", (void *)&kgr_vmx_vpid_lock, "kvm_intel" },
	{ "vmx_vpid_bitmap", (void *)&kgr_vmx_vpid_bitmap, "kvm_intel" },
	{ "host_gdt", (void *)&kgr_host_gdt, "kvm_intel" },
	{ "nested", (void *)&kgr_nested, "kvm_intel" },
	{ "vmx_msr_index", (void *)&kgr_vmx_msr_index, "kvm_intel" },
	{ "copy_shadow_to_vmcs12", (void *)&kgr_copy_shadow_to_vmcs12,
	  "kvm_intel" },
	{ "free_loaded_vmcs", (void *)&kgr_free_loaded_vmcs, "kvm_intel" },
	{ "nested_vmx_check_permission",
	  (void *)&kgr_nested_vmx_check_permission, "kvm_intel" },
	{ "skip_emulated_instruction",
	  (void *)&kgr_skip_emulated_instruction, "kvm_intel" },
	{ "nested_vmx_succeed", (void *)&kgr_nested_vmx_succeed, "kvm_intel" },
	{ "nested_vmx_vmexit", (void *)&kgr_nested_vmx_vmexit, "kvm_intel" },
	{ "vmcs_writel", (void *)&kgr_vmcs_writel, "kvm_intel" },
	{ "guest_cpuid_has_vmx", (void *)&kgr_guest_cpuid_has_vmx,
	  "kvm_intel" },

	{ "kvm_rebooting", (void *)&kgr_kvm_rebooting, "kvm" },
	{ "kvm_vcpu_cache", (void *)&kgr_kvm_vcpu_cache, "kvm" },
	{ "kvm_spurious_fault", (void *)&kgr_kvm_spurious_fault, "kvm" },
	{ "kvm_release_page_dirty", (void *)&kgr_kvm_release_page_dirty,
	  "kvm" },
	{ "kvm_vcpu_uninit", (void *)&kgr_kvm_vcpu_uninit, "kvm" },
	{ "kvm_set_msr_common", (void *)&kgr_kvm_set_msr_common, "kvm" },
	{ "kvm_write_tsc", (void *)&kgr_kvm_write_tsc, "kvm" },
	{ "kvm_mtrr_valid", (void *)&kgr_kvm_mtrr_valid, "kvm" },
	{ "kvm_set_shared_msr", (void *)&kgr_kvm_set_shared_msr, "kvm" },
};


/* from arch/x86/include/asm/kvm_host.h */
/*
 * Avoid module dependency on kvm.ko.
 *
 * This is the original macro but with the load from kvm_rebooting
 * replaced by an indirect load from *kgr_kvm_rebooting and the call
 * to kvm_spurious_fault() replaced by an indirect call to
 * *kgr_kvm_spurious_fault.
 */
#define kgr____kvm_handle_fault_on_reboot(insn, cleanup_insn)	\
	"666: " insn "\n\t" \
	"668: \n\t"                           \
	".pushsection .fixup, \"ax\" \n" \
	"667: \n\t" \
	cleanup_insn "\n\t"		      \
	"pushq %%rax \n\t"		      \
	"movq kgr_kvm_rebooting, %%rax\n\t"   \
	"cmpb $0, (%%rax) \n\t"	      \
	"popq %%rax \n\t"		      \
	"jne 668b \n\t"      		      \
	__ASM_SIZE(push) " $666b \n\t"	      \
	/* this depends on [thunk_target] "a" constraints */	\
	"movq kgr_kvm_spurious_fault, %%rax \n\t"		\
	CALL_NOSPEC "\n\t"					\
	".popsection \n\t" \
	_ASM_EXTABLE(666b, 667b)


/* from arch/x86/kvm/kvm_cache_regs.h */
/* inlined */
static inline bool kgr_is_guest_mode(struct kvm_vcpu *vcpu)
{
	return vcpu->arch.hflags & HF_GUEST_MASK;
}


/* from arch/x86/kvm/x86.h */
/* inlined */
static inline int kgr_is_long_mode(struct kvm_vcpu *vcpu)
{
#ifdef CONFIG_X86_64
	return vcpu->arch.efer & EFER_LMA;
#else
	return 0;
#endif
}


/* from arch/x86/kvm/vmx.c */
#define kgr__ex_clear(x, reg) \
	kgr____kvm_handle_fault_on_reboot(x, "xor " reg " , " reg)

#define KGR_NR_AUTOLOAD_MSRS 8

struct loaded_vmcs {
	struct vmcs *vmcs;
	int cpu;
	int launched;
	struct list_head loaded_vmcss_on_cpu_link;
};

struct shared_msr_entry {
	unsigned index;
	u64 data;
	u64 mask;
};

struct vmcs02_list {
	struct list_head list;
	gpa_t vmptr;
	struct loaded_vmcs vmcs02;
};

struct nested_vmx {
	/* Has the level1 guest done vmxon? */
	bool vmxon;

	/* The guest-physical address of the current VMCS L1 keeps for L2 */
	gpa_t current_vmptr;
	/* The host-usable pointer to the above */
	struct page *current_vmcs12_page;
	struct vmcs12 *current_vmcs12;
	struct vmcs *current_shadow_vmcs;
	/*
	 * Indicates if the shadow vmcs must be updated with the
	 * data hold by vmcs12
	 */
	bool sync_shadow_vmcs;

	/* vmcs02_list cache of VMCSs recently used to run L2 guests */
	struct list_head vmcs02_pool;
	int vmcs02_num;
	u64 vmcs01_tsc_offset;
	/* L2 must run next, and mustn't decide to exit to L1. */
	bool nested_run_pending;
	/*
	 * Guest pages referred to in vmcs02 with host-physical pointers, so
	 * we must keep them pinned while L2 runs.
	 */
	struct page *apic_access_page;
	u64 msr_ia32_feature_control;

	struct hrtimer preemption_timer;
	bool preemption_timer_expired;
};

struct pi_desc {
	u32 pir[8];     /* Posted interrupt requested */
	u32 control;	/* bit 0 of control is outstanding notification bit */
	u32 rsvd[7];
} __aligned(64);

struct vcpu_vmx {
	struct kvm_vcpu       vcpu;
	unsigned long         host_rsp;
	u8                    fail;
	u8                    cpl;
	bool                  nmi_known_unmasked;
	u32                   exit_intr_info;
	u32                   idt_vectoring_info;
	ulong                 rflags;
	struct shared_msr_entry *guest_msrs;
	int                   nmsrs;
	int                   save_nmsrs;
	unsigned long	      host_idt_base;
#ifdef CONFIG_X86_64
	u64 		      msr_host_kernel_gs_base;
	u64 		      msr_guest_kernel_gs_base;
#endif
	u32 vm_entry_controls_shadow;
	u32 vm_exit_controls_shadow;
	/*
	 * loaded_vmcs points to the VMCS currently used in this vcpu. For a
	 * non-nested (L1) guest, it always points to vmcs01. For a nested
	 * guest (L2), it points to a different VMCS.
	 */
	struct loaded_vmcs    vmcs01;
	struct loaded_vmcs   *loaded_vmcs;
	bool                  __launched; /* temporary, used in vmx_vcpu_run */
	struct msr_autoload {
		unsigned nr;
		struct vmx_msr_entry guest[KGR_NR_AUTOLOAD_MSRS];
		struct vmx_msr_entry host[KGR_NR_AUTOLOAD_MSRS];
	} msr_autoload;
	struct {
		int           loaded;
		u16           fs_sel, gs_sel, ldt_sel;
#ifdef CONFIG_X86_64
		u16           ds_sel, es_sel;
#endif
		int           gs_ldt_reload_needed;
		int           fs_reload_needed;
		unsigned long vmcs_host_cr4;    /* May not match real cr4 */
	} host_state;
	struct {
		int vm86_active;
		ulong save_rflags;
		struct kvm_segment segs[8];
	} rmode;
	struct {
		u32 bitmask; /* 4 bits per segment (1 bit per field) */
		struct kvm_save_segment {
			u16 selector;
			unsigned long base;
			u32 limit;
			u32 ar;
		} seg[8];
	} segment_cache;
	int vpid;
	bool emulation_required;

	/* Support for vnmi-less CPUs */
	int soft_vnmi_blocked;
	ktime_t entry_time;
	s64 vnmi_blocked_time;
	u32 exit_reason;

	bool rdtscp_enabled;

	/* Posted interrupt descriptor */
	struct pi_desc pi_desc;

	/* Support for a guest hypervisor (nested VMX) */
	struct nested_vmx nested;

	u64 spec_ctrl;
};


/* inlined */
static inline struct vcpu_vmx *kgr_to_vmx(struct kvm_vcpu *vcpu)
{
	return container_of(vcpu, struct vcpu_vmx, vcpu);
}

/* inlined */
static void kgr_nested_release_page(struct page *page)
{
	kgr_kvm_release_page_dirty(page);
}

struct vmcs_config {
	int size;
	int order;
	u32 revision_id;
	u32 pin_based_exec_ctrl;
	u32 cpu_based_exec_ctrl;
	u32 cpu_based_2nd_exec_ctrl;
	u32 vmexit_ctrl;
	u32 vmentry_ctrl;
};

/* inlined */
static int kgr__find_msr_index(struct vcpu_vmx *vmx, u32 msr)
{
	int i;

	for (i = 0; i < vmx->nmsrs; ++i)
		if ((*kgr_vmx_msr_index)[vmx->guest_msrs[i].index] == msr)
			return i;
	return -1;
}

/* inlined */
static struct shared_msr_entry *kgr_find_msr_entry(struct vcpu_vmx *vmx, u32 msr)
{
	int i;

	i = kgr__find_msr_index(vmx, msr);
	if (i >= 0)
		return &vmx->guest_msrs[i];
	return NULL;
}

/* inlined */
static __always_inline unsigned long kgr_vmcs_readl(unsigned long field)
{
	unsigned long value;

	asm volatile (kgr__ex_clear(ASM_VMX_VMREAD_RDX_RAX, "%0")
		      : [thunk_target] "=a"(value) : "d"(field) : "cc");
	return value;
}

/* inlined */
static __always_inline u32 kgr_vmcs_read32(unsigned long field)
{
	return kgr_vmcs_readl(field);
}

/* inlined */
static void kgr_vmcs_write32(unsigned long field, u32 value)
{
	kgr_vmcs_writel(field, value);
}

/* inlined */
static void kgr_vmcs_write64(unsigned long field, u64 value)
{
	kgr_vmcs_writel(field, value);
#ifndef CONFIG_X86_64
	asm volatile ("");
	kgr_vmcs_writel(field+1, value >> 32);
#endif
}

/* inlined */
static void kgr_vmx_segment_cache_clear(struct vcpu_vmx *vmx)
{
	vmx->segment_cache.bitmask = 0;
}

/* inlined */
static void kgr_reload_tss(void)
{
	/*
	 * VT restores TR but not its size.  Useless.
	 */
	struct desc_ptr *gdt = &__get_cpu_var((*kgr_host_gdt));
	struct desc_struct *descs;

	descs = (void *)gdt->address;
	descs[GDT_ENTRY_TSS].type = 9; /* available TSS */
	load_TR_desc();
}

/* inlined */
static void kgr__vmx_load_host_state(struct vcpu_vmx *vmx)
{
	if (!vmx->host_state.loaded)
		return;

	++vmx->vcpu.stat.host_state_reload;
	vmx->host_state.loaded = 0;
#ifdef CONFIG_X86_64
	if (kgr_is_long_mode(&vmx->vcpu))
		rdmsrl(MSR_KERNEL_GS_BASE, vmx->msr_guest_kernel_gs_base);
#endif
	if (vmx->host_state.gs_ldt_reload_needed) {
		kvm_load_ldt(vmx->host_state.ldt_sel);
#ifdef CONFIG_X86_64
		load_gs_index(vmx->host_state.gs_sel);
#else
		loadsegment(gs, vmx->host_state.gs_sel);
#endif
	}
	if (vmx->host_state.fs_reload_needed)
		loadsegment(fs, vmx->host_state.fs_sel);
#ifdef CONFIG_X86_64
	if (unlikely(vmx->host_state.ds_sel | vmx->host_state.es_sel)) {
		loadsegment(ds, vmx->host_state.ds_sel);
		loadsegment(es, vmx->host_state.es_sel);
	}
#endif
	kgr_reload_tss();
#ifdef CONFIG_X86_64
	wrmsrl(MSR_KERNEL_GS_BASE, vmx->msr_host_kernel_gs_base);
#endif
	/*
	 * If the FPU is not active (through the host task or
	 * the guest vcpu), then restore the cr0.TS bit.
	 */
	if (!user_has_fpu() && !vmx->vcpu.guest_fpu_loaded)
		stts();
	load_gdt(&__get_cpu_var((*kgr_host_gdt)));
}

/* inlined */
static void kgr_vmx_load_host_state(struct vcpu_vmx *vmx)
{
	preempt_disable();
	kgr__vmx_load_host_state(vmx);
	preempt_enable();
}

/* inlined */
static inline bool kgr_nested_vmx_allowed(struct kvm_vcpu *vcpu)
{
	return (*kgr_nested) && kgr_guest_cpuid_has_vmx(vcpu);
}

/* inlined */
static void kgr_free_vmcs(struct vmcs *vmcs)
{
	free_pages((unsigned long)vmcs, (*kgr_vmcs_config).order);
}

/* inlined */
static void kgr_free_vpid(struct vcpu_vmx *vmx)
{
	if (!(*kgr_enable_vpid))
		return;
	spin_lock(&(*kgr_vmx_vpid_lock));
	if (vmx->vpid != 0)
		__clear_bit(vmx->vpid, (*kgr_vmx_vpid_bitmap));
	spin_unlock(&(*kgr_vmx_vpid_lock));
}

/* inlined */
static void kgr_nested_free_all_saved_vmcss(struct vcpu_vmx *vmx)
{
	struct vmcs02_list *item, *n;
	list_for_each_entry_safe(item, n, &vmx->nested.vmcs02_pool, list) {
		if (vmx->loaded_vmcs != &item->vmcs02)
			kgr_free_loaded_vmcs(&item->vmcs02);
		list_del(&item->list);
		kfree(item);
	}
	vmx->nested.vmcs02_num = 0;

	if (vmx->loaded_vmcs != &vmx->vmcs01)
		kgr_free_loaded_vmcs(&vmx->vmcs01);
}

/* inlined */
static inline void kgr_nested_release_vmcs12(struct vcpu_vmx *vmx)
{
	u32 exec_control;
	if ((*kgr_enable_shadow_vmcs)) {
		if (vmx->nested.current_vmcs12 != NULL) {
			/* copy to memory all shadowed fields in case
			   they were modified */
			kgr_copy_shadow_to_vmcs12(vmx);
			vmx->nested.sync_shadow_vmcs = false;
			exec_control = kgr_vmcs_read32(SECONDARY_VM_EXEC_CONTROL);
			exec_control &= ~SECONDARY_EXEC_SHADOW_VMCS;
			kgr_vmcs_write32(SECONDARY_VM_EXEC_CONTROL, exec_control);
			kgr_vmcs_write64(VMCS_LINK_POINTER, -1ull);
		}
	}
	kunmap(vmx->nested.current_vmcs12_page);
	kgr_nested_release_page(vmx->nested.current_vmcs12_page);
}



/* patched, inlined */
static void kgr_free_nested(struct vcpu_vmx *vmx)
{
	if (!vmx->nested.vmxon)
		return;
	/*
	 * Fix CVE-2019-7221
	 *  +1 line
	 */
	hrtimer_cancel(&vmx->nested.preemption_timer);
	vmx->nested.vmxon = false;
	if (vmx->nested.current_vmptr != -1ull) {
		kgr_nested_release_vmcs12(vmx);
		vmx->nested.current_vmptr = -1ull;
		vmx->nested.current_vmcs12 = NULL;
	}
	if ((*kgr_enable_shadow_vmcs))
		kgr_free_vmcs(vmx->nested.current_shadow_vmcs);
	/* Unpin physical memory we referred to in current vmcs02 */
	if (vmx->nested.apic_access_page) {
		kgr_nested_release_page(vmx->nested.apic_access_page);
		vmx->nested.apic_access_page = 0;
	}

	kgr_nested_free_all_saved_vmcss(vmx);
}

/* patched, calls inlined free_nested() */
int kgr_handle_vmoff(struct kvm_vcpu *vcpu)
{
	if (!kgr_nested_vmx_check_permission(vcpu))
		return 1;
	kgr_free_nested(kgr_to_vmx(vcpu));
	kgr_skip_emulated_instruction(vcpu);
	kgr_nested_vmx_succeed(vcpu);
	return 1;
}

/* patched, calls inlined free_nested() */
void kgr_vmx_free_vcpu(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);

	kgr_free_vpid(vmx);
	kgr_free_loaded_vmcs(vmx->loaded_vmcs);
	kgr_free_nested(vmx);
	kfree(vmx->guest_msrs);
	kgr_kvm_vcpu_uninit(vcpu);
	kmem_cache_free((*kgr_kvm_vcpu_cache), vmx);
}

/* patched, inlined, calls inlined free_nested() */
static void kgr_vmx_leave_nested(struct kvm_vcpu *vcpu)
{
	if (kgr_is_guest_mode(vcpu))
		kgr_nested_vmx_vmexit(vcpu, -1, 0, 0);
	kgr_free_nested(kgr_to_vmx(vcpu));
}

/* patched, calls inlined vmx_leave_nested() */
int kgr_vmx_set_msr(struct kvm_vcpu *vcpu, struct msr_data *msr_info)
{
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);
	struct shared_msr_entry *msr;
	int ret = 0;
	u32 msr_index = msr_info->index;
	u64 data = msr_info->data;

	switch (msr_index) {
	case MSR_EFER:
		ret = kgr_kvm_set_msr_common(vcpu, msr_info);
		break;
#ifdef CONFIG_X86_64
	case MSR_FS_BASE:
		kgr_vmx_segment_cache_clear(vmx);
		kgr_vmcs_writel(GUEST_FS_BASE, data);
		break;
	case MSR_GS_BASE:
		kgr_vmx_segment_cache_clear(vmx);
		kgr_vmcs_writel(GUEST_GS_BASE, data);
		break;
	case MSR_KERNEL_GS_BASE:
		kgr_vmx_load_host_state(vmx);
		vmx->msr_guest_kernel_gs_base = data;
		break;
#endif
	case MSR_IA32_SYSENTER_CS:
		kgr_vmcs_write32(GUEST_SYSENTER_CS, data);
		break;
	case MSR_IA32_SYSENTER_EIP:
		kgr_vmcs_writel(GUEST_SYSENTER_EIP, data);
		break;
	case MSR_IA32_SYSENTER_ESP:
		kgr_vmcs_writel(GUEST_SYSENTER_ESP, data);
		break;
	case MSR_IA32_TSC:
		kgr_kvm_write_tsc(vcpu, msr_info);
		break;
	case MSR_IA32_SPEC_CTRL:
		vmx->spec_ctrl = msr_info->data;
		break;
	case MSR_IA32_CR_PAT:
		if ((*kgr_vmcs_config).vmentry_ctrl & VM_ENTRY_LOAD_IA32_PAT) {
			if (!kgr_kvm_mtrr_valid(vcpu, MSR_IA32_CR_PAT, data))
				return 1;
			kgr_vmcs_write64(GUEST_IA32_PAT, data);
			vcpu->arch.pat = data;
			break;
		}
		ret = kgr_kvm_set_msr_common(vcpu, msr_info);
		break;
	case MSR_IA32_TSC_ADJUST:
		ret = kgr_kvm_set_msr_common(vcpu, msr_info);
		break;
	case MSR_IA32_FEATURE_CONTROL:
		if (!kgr_nested_vmx_allowed(vcpu) ||
		    (kgr_to_vmx(vcpu)->nested.msr_ia32_feature_control &
		     FEATURE_CONTROL_LOCKED && !msr_info->host_initiated))
			return 1;
		vmx->nested.msr_ia32_feature_control = data;
		if (msr_info->host_initiated && data == 0)
			kgr_vmx_leave_nested(vcpu);
		break;
	case MSR_IA32_VMX_BASIC ... MSR_IA32_VMX_VMFUNC:
		return 1; /* they are read-only */
	case MSR_TSC_AUX:
		if (!vmx->rdtscp_enabled)
			return 1;
		/* Check reserved bit, higher 32 bits should be zero */
		if ((data >> 32) != 0)
			return 1;
		/* Otherwise falls through */
	default:
		msr = kgr_find_msr_entry(vmx, msr_index);
		if (msr) {
			u64 old_msr_data = msr->data;
			msr->data = data;
			if (msr - vmx->guest_msrs < vmx->save_nmsrs) {
				preempt_disable();
				ret = kgr_kvm_set_shared_msr(msr->index, msr->data,
							     msr->mask);
				preempt_enable();
				if (ret)
					msr->data = old_msr_data;
			}
			break;
		}
		ret = kgr_kvm_set_msr_common(vcpu, msr_info);
	}

	return ret;
}



static int kgr_patch_bsc1124734_module_notify(struct notifier_block *nb,
					      unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = __kgr_resolve_kallsyms_relocs(kgr_funcs, ARRAY_SIZE(kgr_funcs));
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1124734_module_nb = {
	.notifier_call = kgr_patch_bsc1124734_module_notify,
	.priority = INT_MIN+1,
};

int kgr_patch_bsc1124734_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = __kgr_resolve_kallsyms_relocs(kgr_funcs,
						    ARRAY_SIZE(kgr_funcs));
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1124734_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void kgr_patch_bsc1124734_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1124734_module_nb);
}

#endif /* IS_ENABLED(CONFIG_KVM_INTEL) */
